classdef ristra12_input_fields < dynamicprops & ristra12_inputs
    % ristra12_input_fields is a class that extends the ristra12_inputs class,
    % it has additional properties to allow for a user to provide electric
    % fields of their creation using the properties named:
    % 
    % red1_xyt_field_mirror1, red1_xyt_field_mirror2, red2_xyt_field_mirror1,
    % red2_xyt_field_mirror2, and blue_xyt_field_mirror1. The units of the
    % fields are nominally V/m, but their amplitude will be scaled to contain
    % the energy or power specified in properties named
    % ristra_12_left_energy_pwr and ristra_12_right_energy_pwr (inherited from
    % the ristra12_inputs class so not defined here in this classdef file, but
    % can be externally handled as though it is -- setting and getting as
    % normal). The temporal and spatial grids upon which the electric field
    % distributions are defined are also required, and are specified with the
    % properties named xgridmat, ygridmat, tvec_M1, and tvec_M2. The units of
    % xgridmat and ygridmat are meters, and they must be vectors or meshgrid
    % matrices (where the xgridmat property in the 1st dimension is changing &
    % 2nd dimension is constant; ygridmat is constant in the 1st dimension and
    % changing in the 2nd. The spatial grids must also have the same shape and
    % extent - they must be symmetric to keep the ristra12 numerical model
    % simple when it includes the 90 degree rotation of resonated light on each
    % cavity pass.
    % 
    % To provide the user-defined input electric field profiles, set the
    % values of the properties red1_xyt_field_mirror1, red1_xyt_field_mirror2,
    % red2_xyt_field_mirror1, red2_xyt_field_mirror2, blue_xyt_field_mirror1,
    % which are defined on the spatial grids with property names of xgridmat,
    % ygridmat, and temporal grids with property names of tvec_M1 and tvec_M2.
    % 
    % If the input argument to ristra12 is an object of class
    % ristra12_input_fields, the property values from ristra12_inputs defining
    % the spatial and temporal grids, as well as pulse profiles and spatial
    % profiles, are ignored. These properties are ristra_12_nxy,
    % ristra_12_lxy, ristra_12_pulse_duration, ristra_12_temporal_sg,
    % ristra_12_pulse_delay, ristra_12_beam_diameters,
    % ristra_12_spatial_supergaussian, ristra_12_blue_beam_offset_xy,
    % ristra_12_beam_roc.
    % 
    % Also, because the round trip time (and time step) is determined by the
    % cavity and crystal lengths (and refractive indices), the properties
    % ristra_12_cavity_length ristra_12_crystal_lengths ristra_12_ref_inds.
    % 
    % Class methods 
    
    properties
        xgridmat                (:,:)   double;     % x grid positions (meters); should be a vector, or a square 2D array from meshgrid, changing in the 1st dimension and constant in the 2nd dimension
        ygridmat                (:,:)   double;     % y grid positions (meters); should be a vector, or a square 2D array from meshgrid, constant in the 1st dimension and changing in the 2nd dimension
        tvec_M1                 (1,:)   double;     % Temporal grid values (seconds) at mirror M1 inputs; should be a vector, and have the same size as tvec_M2. The vector should be equally spaced.
        tvec_M2                 (1,:)   double;     % Temporal grid values (seconds) at mirror M2 inputs; should be a vector, and have the same size as tvec_M1. The vector should be equally spaced. These values should be delayed from the tvec_M1 values by the time of flight between mirrors M1 and M2.
        red1_xyt_field_mirror1  (:,:,:) double;     % Electric field (V/m) at mirror 1 (left mirror)  for red1 wave; should be a three-dimensional array in x, y, and t (positions and time specified by xgridmat, ygridmat, and tvec_M1)
        red1_xyt_field_mirror2  (:,:,:) double;     % Electric field (V/m) at mirror 2 (right mirror) for red1 wave; should be a three-dimensional array in x, y, and t (positions and time specified by xgridmat, ygridmat, and tvec_M2)
        red2_xyt_field_mirror1  (:,:,:) double;     % Electric field (V/m) at mirror 1 (left mirror)  for red2 wave; should be a three-dimensional array in x, y, and t (positions and time specified by xgridmat, ygridmat, and tvec_M1)
        red2_xyt_field_mirror2  (:,:,:) double;     % Electric field (V/m) at mirror 2 (right mirror) for red2 wave; should be a three-dimensional array in x, y, and t (positions and time specified by xgridmat, ygridmat, and tvec_M2)
        blue_xyt_field_mirror1  (:,:,:) double;     % Electric field (V/m) at mirror 1 (left mirror)  for blue wave; should be a three-dimensional array in x, y, and t (positions and time specified by xgridmat, ygridmat, and tvec_M1)
    end
    methods
        % function o = ristra12_input_fields
        %     % Constructor method, but nothing happens here
        %     if nargout == 0
        %         clear o
        %     end
        % end

        function varargout = make_time_vectors(obj)
            % Using the properties ristra_12_pulse_duration, ristra_12_pulse_delay,
            % ristra_12_nstarts, ristra_12_ref_inds, ristra_12_crystal_lengths,

            c = 3e8; % speed of light m/s
            roundtt     = roundtrip_time(obj);
            durations   = obj.ristra_12_pulse_duration * 1e-9;   % ns -> s
            delays      = [obj.ristra_12_pulse_delay; 0] * 1e-9; % ns -> s
            nstarts     = obj.ristra_12_nstarts;
            n           = obj.ristra_12_ref_inds;
            lcrystals   = obj.ristra_12_crystal_lengths * 1e-3; % mm -> m 

            % Get lengths of cavity legs 1 and 2; cavity leg 1 is between the input and output mirrors, passing through crystal 1; cavity leg 2 is the other leg, from mirror 2 to mirror 3, through crystal 2 (if present), to mirror 4, and back to mirror 1
            [leg_length1, leg_length2] = obj.leg_lengths;

            % cw                  = durations == 0;
            duration_coefficient= 1.85;
            t_start             = min(-delays - duration_coefficient*durations);
            t_end               = max( delays + duration_coefficient*durations);
            nt                  = ceil((t_end-t_start)/roundtt); % number of round trips we have to model
            
            tvec_m1_in  = zeros(nstarts*nt,1);
            tvec_m2_in  = zeros(nstarts*nt,1);

            for I = 1:nstarts
                tvec_m1_in(I:nstarts:end)   = min(delays - duration_coefficient*durations) + ((1:nt)-1)*roundtt + (I-1)*roundtt/nstarts; % setting up a time array for the left inputs and outputs
                tvec_m2_in(I:nstarts:end)   = tvec_m1_in(I:nstarts:end) + 1/c*(n(1)*lcrystals(1) + leg_length1 + leg_length1); 
            end

            obj.tvec_M1 = tvec_m1_in;
            obj.tvec_M2 = tvec_m2_in;
            if nargout == 2
                % Return tvec_M1 and tvec_M2
                varargout{1} = tvec_m1_in;
                varargout{2} = tvec_m2_in;
            elseif nargout == 0
                % Do nothing special
            else
                error('Use 0 or 2 output arguments. You used %i.', nargout);
            end
        end

        function make_spatial_grid(obj)
            nxy                         = obj.ristra_12_nxy;
            Lxy                         = obj.ristra_12_lxy*1e-3;
            xvec                        = linspace(-Lxy/2, Lxy/2, nxy);
            yvec = xvec;
            [obj.ygridmat, obj.xgridmat] = meshgrid(yvec, xvec);
        end
        
        function validate(obj)
            % properties xgridmat and ygridmat must be same in shape, and have the same range
            % assert(isequal(size(obj.xgridmat),size(obj.ygridmat)),'Properties xgridmat and ygridmat must have identical shapes');
            
            p = properties(obj);

            for I = 1:length(p)
                assert(~isempty(obj.(p{I})), ['Property ', (p{I}), ' is empty']);
            end

            assert(isequal(obj.xgridmat, transpose(obj.ygridmat)), 'Properties xgridmat and ygridmat must have identical shapes and range; i.e. xgridmat must be transpose of ygridmat');
            assert(issorted(obj.tvec_M1), 'tvec_M1 is not sorted');
            assert(issorted(obj.tvec_M2), 'tvec_M2 is not sorted');
            assert(issorted(obj.xgridmat(:,1)), 'xgridmat is not sorted');
            assert(issorted(obj.ygridmat(1,:)), 'ygridmat is not sorted');
            
            nt   = length(obj.tvec_M1);
            assert(nt == length(obj.tvec_M2), 'tvec_M2 does not have the same length (%i) as tvec_M1 (%i)', length(obj.tvec_M2), length(obj.tvec_M1));

            % Make sure 1st and 2nd dimensions are Nxy pts long
            npts = size(obj.xgridmat,1);
            assert(size(obj.red1_xyt_field_mirror1,1) == npts, 'red1_xyt_field_mirror1 must have length %i in the 1st dimension; it is length %i instead',  npts, size(obj.red1_xyt_field_mirror1,1));
            assert(size(obj.red1_xyt_field_mirror1,2) == npts, 'red1_xyt_field_mirror1 must have length %i in the 2nd dimension; it is length %i instead',  npts, size(obj.red1_xyt_field_mirror1,2));
            assert(size(obj.red1_xyt_field_mirror2,1) == npts, 'red1_xyt_field_mirror2 must have length %i in the 1st dimension; it is length %i instead',  npts, size(obj.red1_xyt_field_mirror2,1));
            assert(size(obj.red1_xyt_field_mirror2,2) == npts, 'red1_xyt_field_mirror2 must have length %i in the 2nd dimension; it is length %i instead',  npts, size(obj.red1_xyt_field_mirror2,2));
            assert(size(obj.red2_xyt_field_mirror1,1) == npts, 'red2_xyt_field_mirror1 must have length %i in the 1st dimension; it is length %i instead',  npts, size(obj.red2_xyt_field_mirror1,1));
            assert(size(obj.red2_xyt_field_mirror1,2) == npts, 'red2_xyt_field_mirror1 must have length %i in the 2nd dimension; it is length %i instead',  npts, size(obj.red2_xyt_field_mirror1,2));
            assert(size(obj.red2_xyt_field_mirror2,1) == npts, 'red2_xyt_field_mirror2 must have length %i in the 1st dimension; it is length %i instead',  npts, size(obj.red2_xyt_field_mirror2,1));
            assert(size(obj.red2_xyt_field_mirror2,2) == npts, 'red2_xyt_field_mirror2 must have length %i in the 2nd dimension; it is length %i instead',  npts, size(obj.red2_xyt_field_mirror2,2));
            assert(size(obj.blue_xyt_field_mirror1,1) == npts, 'blue_xyt_field_mirror1 must have length %i in the 1st dimension; it is length %i instead',  npts, size(obj.blue_xyt_field_mirror1,1));
            assert(size(obj.blue_xyt_field_mirror1,2) == npts, 'blue_xyt_field_mirror1 must have length %i in the 2nd dimension; it is length %i instead',  npts, size(obj.blue_xyt_field_mirror1,2));

            % Make sure 3rd dimension is length 1 or nt
            assert((size(obj.red1_xyt_field_mirror1,3) == nt) || (size(obj.red1_xyt_field_mirror1,3) == 1), 'The size of red1_xyt_field_mirror1 in the third dimension (%i) must be 1 or the same length as tvec_M1 and tvec_M2 (%i).', size(obj.red1_xyt_field_mirror1,3), nt);
            assert((size(obj.red1_xyt_field_mirror2,3) == nt) || (size(obj.red1_xyt_field_mirror2,3) == 1), 'The size of red1_xyt_field_mirror2 in the third dimension (%i) must be 1 or the same length as tvec_M1 and tvec_M2 (%i).', size(obj.red1_xyt_field_mirror2,3), nt);
            assert((size(obj.red2_xyt_field_mirror1,3) == nt) || (size(obj.red2_xyt_field_mirror1,3) == 1), 'The size of red2_xyt_field_mirror1 in the third dimension (%i) must be 1 or the same length as tvec_M1 and tvec_M2 (%i).', size(obj.red2_xyt_field_mirror1,3), nt);
            assert((size(obj.red2_xyt_field_mirror2,3) == nt) || (size(obj.red2_xyt_field_mirror2,3) == 1), 'The size of red2_xyt_field_mirror2 in the third dimension (%i) must be 1 or the same length as tvec_M1 and tvec_M2 (%i).', size(obj.red2_xyt_field_mirror2,3), nt);
            assert((size(obj.blue_xyt_field_mirror1,3) == nt) || (size(obj.blue_xyt_field_mirror1,3) == 1), 'The size of blue_xyt_field_mirror1 in the third dimension (%i) must be 1 or the same length as tvec_M1 and tvec_M2 (%i).', size(obj.blue_xyt_field_mirror1,3), nt);

            % assert(size(obj.red1_xyt_field_mirror2,3) == nt, 'The size of red1_xyt_field_mirror2 in the third dimension must be the same length as tvec_M1 and tvec_M2.');
            % assert(size(obj.red2_xyt_field_mirror1,3) == nt, 'The size of red2_xyt_field_mirror1 in the third dimension must be the same length as tvec_M1 and tvec_M2.');
            % assert(size(obj.red2_xyt_field_mirror2,3) == nt, 'The size of red2_xyt_field_mirror2 in the third dimension must be the same length as tvec_M1 and tvec_M2.');
            % assert(size(obj.blue_xyt_field_mirror1,3) == nt, 'The size of blue_xyt_field_mirror1 in the third dimension must be the same length as tvec_M1 and tvec_M2.');
            
            % strs = {'1st', '2nd'};
            % for I = 1:length(p)
            %     assert(all(  ~isnan(obj.(p{I})), "all"), [p{I}, ' has nan elements']);
            %     assert(all(isfinite(obj.(p{I})), "all"), [p{I}, ' has is not all finite elements']);
            %     % % What does this bit do? 
            %     % if ~(strcmpi(p{I}, 'tvec_M2') || strcmpi(p{I}, 'tvec_M1')) % if current property name is not tvec_M1 or tvec_M2,
            %     % % if ndims(obj.(p{I})) >= 2
            %     %     for K = 1:2
            %     %         % Make sure the property value in first and second dimensions is same length as spatial grid (size(obj.xgridmat,1))
            %     %         assert(size(obj.(p{I}),K) == npts, ['The length of ', p{I}, ' in the ', strs{K}, ' dimension must have npts (', num2str(npts),') elements']);
            %     %     end
            %     % end
            % end

            % ny = size(obj.xgridmat,2);
            % assert(nx == ny);
            % for K = 1:2
            %     assert(size(obj.red1_xyt_field_mirror1,K) == nx);
            %     assert(size(obj.red1_xyt_field_mirror2,K) == nx);
            %     assert(size(obj.red2_xyt_field_mirror1,K) == nx);
            %     assert(size(obj.red2_xyt_field_mirror2,K) == nx);
            %     assert(size(obj.blue_xyt_field_mirror1,K) == nx);
            % end
            
            % assert(all(~isnan(obj.red1_xyt_field_mirror1),"all"));
            % assert(all(~isnan(obj.red1_xyt_field_mirror2),"all"));
            % assert(all(~isnan(obj.red2_xyt_field_mirror1),"all"));
            % assert(all(~isnan(obj.red2_xyt_field_mirror2),"all"));
            % assert(all(~isnan(obj.blue_xyt_field_mirror1),"all"));
            % assert(all(isfinite(obj.red1_xyt_field_mirror1),"all"));
            % assert(all(isfinite(obj.red1_xyt_field_mirror2),"all"));
            % assert(all(isfinite(obj.red2_xyt_field_mirror1),"all"));
            % assert(all(isfinite(obj.red2_xyt_field_mirror2),"all"));
            % assert(all(isfinite(obj.blue_xyt_field_mirror1),"all"));
        end

        function make_gaussian_pulses(obj)
            % Generates vector of pulse powers for each wave
            %
            % Using the properties tvec_M1, tvec_M2, ristra_12_pulse_duration,
            % ristra_12_pulse_delay, ristra_12_temporal_sg. tvec_M1 and tvec_M2 can be defined by
            % the user, or calling the method make_time_vectors will generate these vectors using
            % properties such as ristra_12_crystal_lengths, ristra_12_cavity_length,
            % ristra_12_ref_inds, and ristra_12_nstarts. Pulse profiles will be scaled when the
            % ristra12 model run button is called.
            %
            % For cw beams, this method sets the pulse profiles to be a vector of ones (scaled when
            % the run function is called).
            %

            if isempty(obj.tvec_M1) || isempty(obj.tvec_M2)
                obj.make_time_vectors;
            end

            tvecM1      = obj.tvec_M1;
            tvecM2      = obj.tvec_M2;
            durations   =  obj.ristra_12_pulse_duration * 1e-9;
            delays      = [obj.ristra_12_pulse_delay;0] * 1e-9;
            superg_t    = obj.ristra_12_temporal_sg;
            if ~isprop(obj, 'pulse_profiles')
                addprop(obj, 'pulse_profiles');
            end
            if durations(1) ~= 0
                red1_M1 = exp(-log(2)*((2*(tvecM1 - delays(1))/durations(1)).^2).^superg_t(1)); % nt x 1 array
                red1_M2 = exp(-log(2)*((2*(tvecM2 - delays(1))/durations(1)).^2).^superg_t(1)); % nt x 1 array
            else
                red1_M1 = ones(size(tvecM1));
                red1_M2 = ones(size(tvecM1));
            end
            if durations(2) ~= 0
                red2_M1 = exp(-log(2)*((2*(tvecM1 - delays(2))/durations(2)).^2).^superg_t(2)); % nt x 1 array
                red2_M2 = exp(-log(2)*((2*(tvecM2 - delays(2))/durations(2)).^2).^superg_t(2)); % nt x 1 array
            else
                red2_M1 = ones(size(tvecM1));
                red2_M2 = ones(size(tvecM1));
            end
            if durations(3) ~= 0
                blue_M1 = exp(-log(2)*((2*(tvecM1 - delays(3))/durations(3)).^2).^superg_t(3)); % nt x 1 array
            else
                blue_M1 = ones(size(tvecM1));
            end

            obj.pulse_profiles.red1_M1 = red1_M1;
            obj.pulse_profiles.red1_M2 = red1_M2;
            obj.pulse_profiles.red2_M1 = red2_M1;
            obj.pulse_profiles.red2_M2 = red2_M2;
            obj.pulse_profiles.blue_M1 = blue_M1;
        end

        function make_Exyt(obj)
            if ~isprop(obj, 'spatial_profiles')
                obj.make_supergaussian_spatial_profiles;
            end
            if ~isprop(obj, 'pulse_profiles')
                obj.make_gaussian_pulses;
            end
            % keyboard;
            nt = length(obj.tvec_M1);
            nx = size(obj.spatial_profiles.M1_field_xy_norm_red1,1);
            ny = nx;
            obj.red1_xyt_field_mirror1 = repmat(obj.spatial_profiles.M1_field_xy_norm_red1,[1,1,nt]) .* repmat(permute(sqrt(obj.pulse_profiles.red1_M1),[3,1,2]),[nx,ny,1]);
            obj.red2_xyt_field_mirror1 = repmat(obj.spatial_profiles.M1_field_xy_norm_red2,[1,1,nt]) .* repmat(permute(sqrt(obj.pulse_profiles.red2_M1),[3,1,2]),[nx,ny,1]);
            obj.blue_xyt_field_mirror1 = repmat(obj.spatial_profiles.M1_field_xy_norm_blue,[1,1,nt]) .* repmat(permute(sqrt(obj.pulse_profiles.blue_M1),[3,1,2]),[nx,ny,1]);
            obj.red1_xyt_field_mirror2 = repmat(obj.spatial_profiles.M2_field_xy_norm_red1,[1,1,nt]) .* repmat(permute(sqrt(obj.pulse_profiles.red1_M2),[3,1,2]),[nx,ny,1]);
            obj.red2_xyt_field_mirror2 = repmat(obj.spatial_profiles.M2_field_xy_norm_red2,[1,1,nt]) .* repmat(permute(sqrt(obj.pulse_profiles.red2_M2),[3,1,2]),[nx,ny,1]);
            % keyboard;
        end

        function make_supergaussian_spatial_profiles(obj)
            % Using the properties ristra_12_wavelengths, ristra_12_beam_diameters,
            % ristra_12_blue_beam_offset_xy, ristra_12_spatial_supergaussian, and
            % ristra_12_beam_roc, generate the supergaussian electric field profiles for
            % each wave at both mirrors, set results to properties fields in a new
            % 'spatial_profiles' property; there, the fieldnames will be
            % M1_field_xy_norm_red1, M1_field_xy_norm_red2, M1_field_xy_norm_blue,
            % M2_field_xy_norm_red1, M2_field_xy_norm_red2.
            %
            % The fields will be normalized to have 1 watt of power, and the radius of
            % curvature added using the values in the property ristra_12_beam_roc.
            
            c               = 3e8;                              % speed of light (m/s)
            epsilon_0       = 8.85e-12;                         % vacuum permittivity (F/m)

            % try to use the properties xgridmat and ygridmat if they already exist, but
            % if they don't run the make_spatial_grid method to generate them:
            if isempty(obj.xgridmat) || isempty(obj.ygridmat)
                obj.make_spatial_grid;
            else
                % They already exist; shall we use their values to set the properties
                % dealing with the # pts and range of x & y?
                obj.ristra_12_lxy = diff([min(obj.xgridmat,[],'all'), max(obj.xgridmat,[],'all')]) * 1e3; % in mm
            end
            
            % % TODO: How shall we validate pulse profiles & time grids?
            % roundtt         = obj.roundtrip_time;

            nx              = length(obj.xgridmat);
            ny              = nx;

            lambda          = obj.ristra_12_wavelengths*1e-9;           % nm -> m
            diameter_x      = obj.ristra_12_beam_diameters*1e-3;        % mm -> m
            diameter_y      = diameter_x;
            xgrid           = obj.xgridmat(:,1);
            ygrid           = xgrid;
            dx              = xgrid(2) - xgrid(1);
            dy              = dx;
            blueoffsets     = obj.ristra_12_blue_beam_offset_xy*1e-3;   % mm -> m
            superg_xy       = obj.ristra_12_spatial_supergaussian;
            roc             = obj.ristra_12_beam_roc;

            %  The energies/powers are not used to scale the electric fields, we'll do
            %  that in ristra12 when the run button is pressed using the property
            %  ristra_12_left_energy_pwr and ristra_12_right_energy_pwr properties:
            % u_M1                    = obj.ristra_12_left_energy_pwr;
            % u_M2                    = obj.ristra_12_right_energy_pwr;

            roc_beam_x              = roc(:,1);
            roc_beam_y              = roc(:,2);

            exponent_x              = zeros(nx,3);
            exponent_x(:,1)         = ( (2./diameter_x(1)).*(xgrid) ).^2; % red 1 spatial profile in x (with proper diameter, always centered)
            exponent_x(:,2)         = ( (2./diameter_x(2)).*(xgrid) ).^2; % red 2 spatial profile in x (with proper diameter, always centered)
            exponent_x(:,3)         = ( (2./diameter_x(3)).*(xgrid - blueoffsets(1)) ).^2; % blue spatial profile in x (with proper diameter and center)

            % set up exponent for gaussian field profiles in y (for each of the three waves)
            exponent_y = zeros(ny,3);
            exponent_y(:,1)         = ( 2./diameter_y(1).*ygrid ).^2; % red 1 spatial profile in y (with proper diameter)
            exponent_y(:,2)         = ( 2./diameter_y(2).*ygrid ).^2; % red 2 spatial profile in y (with proper diameter)
            exponent_y(:,3)         = ( 2./diameter_y(3).*(ygrid - blueoffsets(2)) ).^2; % blue  spatial profile in y (with proper diameter)

            % 2D (x,y) super gaussian irradiances for red1, red2, blue waves: the command
            % repmat(vec1,[1,ny])+repmat(vec2.',[nx,1]) is an outer sum
            left_irrad_xy_red1         = exp( -log(2)*(repmat(exponent_x(:,1),[1,ny])+repmat(exponent_y(:,1).',[nx,1])).^(superg_xy(1)) ); % red1 spatial 2d profile
            left_irrad_xy_red2         = exp( -log(2)*(repmat(exponent_x(:,2),[1,ny])+repmat(exponent_y(:,2).',[nx,1])).^(superg_xy(2)) ); % red2 spatial 2d profile
            left_irrad_xy_blue         = exp( -log(2)*(repmat(exponent_x(:,3),[1,ny])+repmat(exponent_y(:,3).',[nx,1])).^(superg_xy(3)) ); % blue spatial 2d profile

            % normalize the fields to have total of 1 W power each when integrated over dx and dy
            left_irrad_xy_norm_red1     = left_irrad_xy_red1 ./ (dx*dy*sum(sum(left_irrad_xy_red1)));
            left_irrad_xy_norm_red2     = left_irrad_xy_red2 ./ (dx*dy*sum(sum(left_irrad_xy_red2)));
            left_irrad_xy_norm_blue     = left_irrad_xy_blue ./ (dx*dy*sum(sum(left_irrad_xy_blue)));

            %% Make irradiance profiles into electric field profiles
            % They'll be purely real at this point (infinite radius of curvature), but
            % we'll apply spherical curvature below:

            % make normalized transverse fields from normalized transverse irradiances for red1, red2, blue, for
            % inputting from the left side in the loop, we'll multiply these by their power at time of input at mirror 1.

            M1_field_xy_norm_flat_red1  = sqrt( left_irrad_xy_norm_red1 / (epsilon_0*c/2) );
            M1_field_xy_norm_flat_red2  = sqrt( left_irrad_xy_norm_red2 / (epsilon_0*c/2) );
            M1_field_xy_norm_flat_blue  = sqrt( left_irrad_xy_norm_blue / (epsilon_0*c/2) );

            % Now all the fields
            % right input fields have same shape as left for red1 & red2 (there is no right blue
            % input). These values should be set such that integrating them in x and y gives a value
            % of 1 watt; we'll scale them to have the appropriate power at each point in time when we
            % reach the mirrors.
            M2_field_xy_norm_flat_red1  = M1_field_xy_norm_flat_red1;
            M2_field_xy_norm_flat_red2  = M1_field_xy_norm_flat_red2;

            % not used:
            % M1_fields_norm              = cat(3, M1_field_xy_norm_red1, M1_field_xy_norm_red2, M1_field_xy_norm_blue);
            % M2_fields_norm              = cat(3, M1_field_xy_norm_red1, M1_field_xy_norm_red2);

            %% Add the curvatures:
            % calculate curvatures for each wave at mirror inputs (add these to the normalized input fields):
            curvature_m1_beam_red1      = -pi/lambda(1).*(obj.xgridmat.^2./roc_beam_x(1) + obj.ygridmat.^2./roc_beam_y(1));
            curvature_m1_beam_red2      = -pi/lambda(2).*(obj.xgridmat.^2./roc_beam_x(2) + obj.ygridmat.^2./roc_beam_y(2));
            curvature_m1_beam_blue      = -pi/lambda(3).*((obj.xgridmat-blueoffsets(1)).^2./roc_beam_x(3) + (obj.ygridmat-blueoffsets(2)).^2./roc_beam_y(3));
            curvature_m2_beam_red1      = curvature_m1_beam_red1;
            curvature_m2_beam_red2      = curvature_m1_beam_red2;

            M1_field_xy_norm_red1       = M1_field_xy_norm_flat_red1 .* exp(1i*curvature_m1_beam_red1);
            M1_field_xy_norm_red2       = M1_field_xy_norm_flat_red2 .* exp(1i*curvature_m1_beam_red2);
            M1_field_xy_norm_blue       = M1_field_xy_norm_flat_blue .* exp(1i*curvature_m1_beam_blue);
            M2_field_xy_norm_red1       = M2_field_xy_norm_flat_red1 .* exp(1i*curvature_m2_beam_red1);
            M2_field_xy_norm_red2       = M2_field_xy_norm_flat_red2 .* exp(1i*curvature_m2_beam_red2);

            if ~isprop(obj, 'spatial_profiles')
                addprop(obj, 'spatial_profiles');
            end
            obj.spatial_profiles.M1_field_xy_norm_red1 = M1_field_xy_norm_red1;
            obj.spatial_profiles.M1_field_xy_norm_red2 = M1_field_xy_norm_red2;
            obj.spatial_profiles.M1_field_xy_norm_blue = M1_field_xy_norm_blue;
            obj.spatial_profiles.M2_field_xy_norm_red1 = M2_field_xy_norm_red1;
            obj.spatial_profiles.M2_field_xy_norm_red2 = M2_field_xy_norm_red2;
            % Note: still have to combine the pulse profiles with these transverse fields to
            % define the full fields in x, y, and t. Do that with the method make_Exyt.
        end

    end
end
