classdef ristra12_inputs < dynamicprops & matlab.mixin.Copyable
    % A class to hold all of the input parameter values for Ristra 12 model.
    %
    % version 1.2 - October 24, 2024
    %
    % Note that this can be used with subclass ristra12_input_fields, which
    % allows custom, user-specified input electric field profiles. Using that
    % class, the values for several properties of this class may be in conflict
    % with the ristra12_input_fields - namely, nx, Lx, and things like duration,
    % energy, delay, diameters, supergaussian coefficients, radius of
    % curvature).
    % 
    % The model starts with the first point in time at the left mirror (also
    % called mirror 1, or input mirror), enters the cavity by transmitting
    % through mirror 1, propagates to the input face of crystal 1, enters
    % crystal 1, mixes and propagates through crystal 1, exits crystal,
    % propagates to mirror 2, reflect & transmit through mirror 2, through the
    % first half of leg 2 to crystal 2, enters crystal 2, mixes and propagates
    % through crystal 2, exits crystal 2, propagates last half of leg 2 to the
    % input mirror, and reflects & transmits through mirror 1; now step time &
    % move to the next cavity roundtrip.
    % 
    % At both the input and output mirror, electric fields for each wave enter &
    % exit cavity, with the exception of having no input field for the blue wave
    % at the output mirror.
    % 
    % For long cavities or short pulses, the time step (set by the cavity
    % roundtrip time) may be too large, in which case you can use multiple
    % 'starts.'
    %
    % Since each of the properties defined here include the shape the values
    % must take, if you assign a scalar to any of the non-scalar values the
    % scalar will be replicated to have the appropriate shape.
    % 
    % Beams size, profile, offset, and power/energy are specified at the cavity
    % mirrors M1 ('left' some places) and M2 ('right' some places) before
    % entering cavity through that mirror. Beam properties for a wave are shared
    % between M1 and M2 inputs, but are individually scaled to have the power or
    % energy specified in ristra_12_left_energy_pwr or
    % ristra_12_right_energy_pwr.
    % 
    % A description of the output variables of the ristra12 model can be found
    % in Ristra12_output_variables.html
    % 
properties
   	% (3x1 vector) wavelengths [red1, red2, blue] (nm)
    ristra_12_wavelengths        	(3,1) double {mustBeNumeric, mustBeReal, mustBeFinite, mustBePositive} = [890, 1271.3, 523.5]; 
    % (3x1 vector) refractive index at wavelength [red1, red2, blue] 
	ristra_12_ref_inds           	(3,1) double {mustBeNumeric, mustBeReal, mustBeFinite, mustBePositive} = [1.609, 1.601, 1.606];
	% (3x1 vector) (power) reflectivity of crystal input faces (light is lost, not resonated) [red1, red2, blue] (range 0-1)
    ristra_12_left_refl_c        	(3,1) double {mustBeNumeric, mustBeReal, mustBeFinite, mustBeNonnegative, mustBeInRange(ristra_12_left_refl_c,  0, 1, 'exclude-upper')} = [0, 0, 0];    
	% (3x1 vector) (power) reflectivity of crystal output faces (light is lost, not resonated) [red1, red2, blue]  (range 0-1)
    ristra_12_right_refl_c       	(3,1) double {mustBeNumeric, mustBeReal, mustBeFinite, mustBeNonnegative, mustBeInRange(ristra_12_right_refl_c, 0, 1, 'exclude-upper')} = [0, 0, 0];
	% (3x1) vector linear absorption [red1, red2, blue] (/mm)
    ristra_12_crystal_loss       	(3,1) double {mustBeNumeric, mustBeReal, mustBeFinite, mustBeNonnegative} = [0, 0, 0];
	% (3x1 vector) incident power or energy at mirror M1, the left mirror; [red1, red2, blue]  (watts if cw, joules if pulsed)
    ristra_12_left_energy_pwr   	(3,1) double {mustBeNumeric, mustBeReal, mustBeFinite, mustBeNonnegative} = [0, 0, 0.60];                
	% (3x1 vector) incident power or energy at mirror M2, the right mirror; [red1, red2, blue]  (watts if cw, joules if pulsed)
    ristra_12_right_energy_pwr  	(3,1) double {mustBeNumeric, mustBeReal, mustBeFinite, mustBeNonnegative} = [1e-6, 1e-6, 0];                      
	% (3x1 vector) pulse duration (fwhm power) [red1, red2, blue] (ns)
    ristra_12_pulse_duration   		(3,1) double {mustBeNumeric, mustBeReal, mustBeFinite, mustBeNonnegative} = [30, 0, 31];                      
	% (3x1 vector) temporal supergaussian index (values of 1 are Gaussians; larger values are more flat-topped) [red1, red2, blue]
    ristra_12_temporal_sg   		(3,1) double {mustBeNumeric, mustBeReal, mustBeFinite, mustBeNonnegative, mustBeGreaterThanOrEqual(ristra_12_temporal_sg, 1)} = [1, 1, 1];                      
	% (2x1 vector) red1 & red2 pulse delay relative to blue wave [red1, red2, blue] (ns)
    ristra_12_pulse_delay   		(2,1) double {mustBeNumeric, mustBeReal, mustBeFinite} = [0, 0];                                            
	% (3x2 array) beam diameter (mm full-width half-max). Order is [red1 x, red1 y; red2 x, red2 y; blue x, blue y]] (mm)
    ristra_12_beam_diameters   		(3,2) double {mustBeNumeric, mustBeReal, mustBeFinite, mustBeNonnegative} = [1.8, 1.8; 1.8, 1.8; 1.8, 1.8];                      
	% (3x1 vector) spatial supergaussian index (values of 1 are Gaussian; larger values are more flat-topped) [red1, red2, blue]
    ristra_12_spatial_supergaussian (3,1) double {mustBeNumeric, mustBeReal, mustBeFinite, mustBeGreaterThanOrEqual(ristra_12_spatial_supergaussian, 1)} = [1, 1, 1];            
	% (2x1 vector) offset of blue wave in the x / y directions [walkoff direction]/[perpendicular to walkoff direction] (mm)
    ristra_12_blue_beam_offset_xy  	(2,1) double {mustBeNumeric, mustBeReal, mustBeFinite} = [0, 0];                                            
	% (3x1 vector) birefringent walk off angles [red1, red2, blue] (mrad)
    ristra_12_walkoff_angles   		(3,1) double {mustBeNumeric, mustBeReal, mustBeFinite} = [0, 100, 0];                                         
    % (3x2 array) beam radius of curvature in the x / y directions (in the direction of birefringent walkoff, at the input mirror)/(perpendicular to walkoff). Order is: [red1 x, red1 y; red2 x, red2 y; blue x, blue y] (mm)
	ristra_12_beam_roc   			(3,2) double {mustBeNumeric, mustBeReal, mustBeNonzero} = ones(3,2)*1e9;  
	% (3x1 vector) left mirror power reflectivity (at wavelength specified) [red1, red2, blue] (range 0-1)
    ristra_12_left_refl_m   		(3,1) double {mustBeNumeric, mustBeReal, mustBeFinite, mustBeNonnegative, mustBeInRange(ristra_12_left_refl_m, 0, 1)} = [0.95, 0.0, 0.0]; 
    % (3x1 vector) right mirror power reflectivity (at wavelength specified) [red1, red2, blue] (range 0-1)
	ristra_12_right_refl_m   		(3,1) double {mustBeNumeric, mustBeReal, mustBeFinite, mustBeNonnegative, mustBeInRange(ristra_12_right_refl_m, 0, 1)} = [0.70, 0.0, 0.990]; 
	% (3x1 vector) phase shift between mirror 1 (the left mirror) and crystal 1 [red1, red2, blue] (radians)
    ristra_12_phase_1   			(3,1) double {mustBeNumeric, mustBeReal, mustBeFinite} = [0, 0, 0]; 
    % (3x1 vector) phase shift between crystal 1 and mirror 2 (the right mirror) [red1, red2, blue] (radians)
	ristra_12_phase_2   			(3,1) double {mustBeNumeric, mustBeReal, mustBeFinite} = [0, 0, 0]; 
    % (scalar) # z integration & mixing pts through first crystal
	ristra_12_nz   			    	(1,1) double {mustBeNumeric, mustBeReal, mustBeFinite, mustBePositive, mustBeInteger} = 30; 
	% (2x1 vector) length of first and second crystal [crystal 1, crystal 2] (mm)
    ristra_12_crystal_lengths   	(2,1) double {mustBeNumeric, mustBeReal, mustBeFinite, mustBeNonnegative} = [9, 9]; 
	% (scalar) total cavity length (mm)
    ristra_12_cavity_length   		(1,1) double {mustBeNumeric, mustBeReal, mustBeFinite, mustBePositive} = 108; 
    % (scalar) # pts in transverse dimension. Value is used for both x and y directions to satisfy symmetry requirements.
	ristra_12_nxy   				(1,1) double {mustBeNumeric, mustBeReal, mustBeFinite, mustBePositive, mustBeInteger} = 64; 
    % (scalar) length of transverse dimension (mm). Value is used for both x and y directions to satisfy symmetry requirements.
	ristra_12_lxy   				(1,1) double {mustBeNumeric, mustBeReal, mustBeFinite, mustBePositive} = 5.50; 
    % (scalar) effective nonlinear coefficient (pm/V)
	ristra_12_deff   				(1,1) double {mustBeNumeric, mustBeReal, mustBeFinite} = 0.832; 
    % (scalar) phase mismatch (rad/mm)
	ristra_12_deltak   				(1,1) double {mustBeNumeric, mustBeReal, mustBeFinite} = 0; 
    % (scalar) # starts through cavity; try 1 to start, larger values to give better temporal resolution (but is much slower & consumes more memory)
	ristra_12_nstarts 				(1,1) double {mustBeNumeric, mustBeReal, mustBeFinite, mustBePositive} = 1; 
    % (scalar) add quantum noise to red1 and red2 waves? 
    ristra_12_add_quantum_noise     (1,1) logical = true;
    % (scalar) disable the 90 degree rotation of resonated wave on each cavity round trip - just make it a non-inverting planar ring cavity
    ristra_12_disable_image_rotation(1,1) logical = false;
end

% As a reminder, the property definition syntax is:
% <property name> <size> <class> <validators> = <default values>
methods
    function o = ristra12_inputs
        % Constructor method; nothing happens here
        if nargout == 0
            clear('o');
        end
    end
    function roundtt = roundtrip_time(obj)
        % Using the properties ristra_12_ref_inds(1), ristra_12_cavity_length,
        % and ristra_12_crystal_lengths(1) & ristra_12_crystal_lengths(2),
        % calculate the cavity round trip time. This time will be used as the 
        % time step when defining temporal grids.

        c           = 3e8;                              % Speed of light (m/s)
        n           = obj.ristra_12_ref_inds;           % refractive indices vector
        L_cavity    = obj.ristra_12_cavity_length;      % total length of cavity (mm)
        L_crystals  = obj.ristra_12_crystal_lengths;    % lengths of crystal 1 and 2 (mm)
        
        % Convert to mks units:
        L_cavity    = L_cavity   * 1e-3;
        L_crystals  = L_crystals * 1e-3;

        % Leg lengths:
        % distance from mirror M1 to 1st crystal input face, 1st crystal output face to mirror M2
        % and      
        % distance from mirror M2 to 2nd crystal input face, from 2nd crystal output face to mirror M1
        
        [leg1, leg2] = obj.leg_lengths;
        
        % round trip time (s): through both pairs of legs in air & through crystals with ref ind > 1:
        roundtt = 1/c * ( 2*(leg1+leg2) + n(1)*(L_crystals(1)+L_crystals(2)) );
    end
    function [leg1, leg2] = leg_lengths(obj)
        % Calculate 
        %  - 'leg1' which is the distance between cavity mirror M1 and crystal 1
        % input face (which is also the distance between crystal 1 output face
        % and cavity mirror M2).
        %   - 'leg2' which is the distance between cavity mirror M2 and crystal 2
        % input face (which is also the distance between crystal 2 output face
        % and cavity mirror M1).

        L_cavity    = obj.ristra_12_cavity_length;  % total length of cavity (mm)
        L_crystals  = obj.ristra_12_crystal_lengths; % lengths of crystal 1 and 2 (mm)

        % Convert to mks units
        L_cavity    = L_cavity   * 1e-3;
        L_crystals  = L_crystals * 1e-3;
        % distance from mirror M1 to 1st crystal input face, 1st crystal output face to mirror M2
        leg1        =    (L_cavity / (2 + sqrt(2)) - L_crystals(1)) * 0.5;

        % distance from mirror M2 to 2nd crystal input face, from 2nd crystal output face to mirror M1
        leg2        = (L_cavity / (2 + sqrt(2)) - L_crystals(2))  * 0.5 + L_cavity / 2 / (1 + sqrt(2));
    end

    function [tvec_M1, tvec_M2] = time_vectors(obj,duration_coefficient)
        % Make vectors of times the pulse profiles will use. The start and stop
        % time should be sufficiently far enough from the pulse center (at t=0)
        % that the full input pulse will be modeled, and the light in the cavity
        % has decayed to some tiny amount.
        arguments
            obj (1,1)
            duration_coefficient (1,1) double {mustBePositive} = 1.5;
        end
        c                   = 3e8; % speed of light m/s
        durations           = obj.ristra_12_pulse_duration;
        delays              = obj.ristra_12_pulse_delay;
        roundtt             = obj.roundtrip_time;
        lcrystals           = obj.ristra_12_crystal_lengths;
        L_cavity            = obj.ristra_12_cavity_length;
        nstarts             = obj.ristra_12_nstarts;
        n                   = obj.ristra_12_ref_inds;

        [length1, length2]  = obj.leg_lengths;

        delays              = [delays(:); 0];
        durations           = durations*1e-9; % ns -> s
        delays              = delays*1e-9; % ns -> s
        lcrystals           = lcrystals*1E-3;       % mm->m
        L_cavity            = L_cavity*1e-3;        % mm -> m


        % duration_coefficient= 1.55;
        t_start             = min(-delays - duration_coefficient*durations);
        t_end               = max( delays + duration_coefficient*durations);

        nt                  = ceil((t_end-t_start)/roundtt); % number of round trips we have to model
        total_nt            = nt * nstarts;

        tvec_M1             = zeros(total_nt,1);
        tvec_M2             = zeros(total_nt,1);

        for I = 1:nstarts
            tvec_M1(I:nstarts:end)   = min(delays - duration_coefficient*durations) + ((1:nt)-1)*roundtt + (I-1)*roundtt/nstarts; % setting up a time array for the left inputs and outputs
            tvec_M2(I:nstarts:end)   = tvec_M1(I:nstarts:end) + 1/c*(n(1)*lcrystals(1) + length1 + length1); % is this correct?
        end
    end

    function O = make_pulses(obj)
        % keyboard;
        % TODO: does this work?
        % Create vectors of (super)Gaussian pulses of power for each wave at mirrors M1 and M2.
        
        c           = 3e8; % speed of light m/s

        [tvec_M1, tvec_M2] = obj.time_vectors;
        temporal_sg = obj.ristra_12_temporal_sg;
        durations   = obj.ristra_12_pulse_duration;
        delays      = obj.ristra_12_pulse_delay;
        L_energies  = obj.ristra_12_left_energy_pwr;
        R_energies  = obj.ristra_12_right_energy_pwr;
        nt          = length(tvec_M1);
        P_red1      = zeros(nt, 2); % for mirror 1 & mirror 2
        P_red2      = zeros(nt, 2); % for mirror 1 & mirror 2
        P_blue      = zeros(nt, 1); % for mirror 1 & mirror 2
        
        durations   = durations * 1e-9; % ns -> s
        delays      = [delays(:);0]*1e-9; % ns -> s

        cw = durations == 0;
        if cw(1)
            P_red1(:,1)     = L_energies(1);
            P_red1(:,2)     = R_energies(1);
        else
            if L_energies(1)== 0
                P_red1(:,1) = 0;
            else
                P1          = exp(-log(2)*((2*(tvec_M1-delays(1)) / durations(1)).^2).^temporal_sg(1));
                P_red1(:,1) = P1 ./ (trapz(tvec_M1, P1) + 1e-80) * L_energies(1);
            end
            if R_energies(1)== 0
                P_red1(:,2) = 0;
            else
                P2          = exp(-log(2)*((2*(tvec_M2-delays(1)) / durations(1)).^2).^temporal_sg(1));
                P_red1(:,2) = P2 ./ (trapz(tvec_M2, P2) + 1e-80) * R_energies(1);
            end
        end

        if cw(2)
            P_red2(:,1)     = L_energies(2);
            P_red2(:,2)     = R_energies(2);
        else
            if L_energies(2)== 0
                P_red2(:,1) = 0;
            else
                P1          = exp(-log(2)*((2*(tvec_M1-delays(2)) / durations(2)).^2).^temporal_sg(2));
                P_red2(:,1) = P1 ./ (trapz(tvec_M1, P1) + 1e-80) * L_energies(2);
            end
            if R_energies(2)== 0
                P_red2(:,2) = 0;
            else
                P2          = exp(-log(2)*((2*(tvec_M2-delays(2)) / durations(2)).^2).^temporal_sg(2));
                P_red2(:,2) = P2 ./ (trapz(tvec_M2, P2) + 1e-80) * R_energies(2);
            end
        end

        if cw(3)
            P_blue(:,1)     = L_energies(3);
        else
            if L_energies(3)== 0
                P_blue(:,1) = 0;
            else
                P1          = exp(-log(2)*((2*(tvec_M1-delays(3)) / durations(3)).^2).^temporal_sg(3));
                P_blue(:,1) = P1 ./ (trapz(tvec_M1, P1) + 1e-80) * L_energies(3);
            end
        end

        % keyboard;
        O.P_red1 = P_red1;
        O.P_red2 = P_red2;
        O.P_blue = P_blue;
        O.tvec_M1= tvec_M1;
        O.tvec_M2= tvec_M2;
    end
    function [S0, S02] = calculate_S_o(obj)
        % Calculate characteristic irradiance for each crystal.
        
        % physical constants
        h               = 6.626E-34;                                    % planck's
        c               = 3e8;                                          % speed of light (m/s)
        epsilon_0       = 8.85e-12;                                     % vacuum permittivity (F/m)

        n               = obj.ristra_12_ref_inds;
        deff            = obj.ristra_12_deff;
        lambda          = obj.ristra_12_wavelengths;
        lcrystals       = obj.ristra_12_crystal_lengths;
        lambda          = lambda * 1e-9; % nm --> m
        deff            = deff*1E-12;           % pm/V->m/V
        lcrystals       = lcrystals*1e-3;
        omega           = 2*pi*c./lambda;
        if deff == 0
            S0      = 0;
            S02     = 0;
        else
            S0      = epsilon_0*c^3*prod(n)./(2*deff^2*omega(1)*omega(2)*lcrystals(1).^2); % just for 1st crystal
            if lcrystals(2) ~= 0
                S02 = epsilon_0*c^3*prod(n)./(2*deff^2*omega(1)*omega(2)*lcrystals(2).^2); % just for 1st crystal
            else
                S02 = 0;
            end
        end
    end

    % function [I_red1, I_red2, I_blue, I_maxs, xmat, ymat] = make_spatial_supergaussians(obj)
    function O = make_spatial_supergaussians_xyt(obj)
        % TODO: add phase from radius of curvature
        c           = 3e8;
        epsilon_0   = 8.85e-12;
        Lx          = obj.ristra_12_lxy;
        L_energies  = obj.ristra_12_left_energy_pwr;
        R_energies  = obj.ristra_12_right_energy_pwr;
        durations   = obj.ristra_12_pulse_duration;
        nx          = obj.ristra_12_nxy;
        blue_offsets= obj.ristra_12_blue_beam_offset_xy;
        roc         = obj.ristra_12_beam_roc;
        diameters   = obj.ristra_12_beam_diameters;
        sg_coeffs   = obj.ristra_12_spatial_supergaussian;

        Lx          = Lx*1e-3; % mm -> m
        blue_offsets= blue_offsets*1e-3; 
        roc         = roc * 1e-3;
        diameters   = diameters*1e-3;

        %% Time grids
        if any(durations ~= 0)
            % At least one is pulsed: we need time vectors
            [tvec_M1, tvec_M2] = obj.time_vectors;
            nt = length(tvec_M1);
            dt = tvec_M1(2) - tvec_M1(1);
        end
        P = obj.make_pulses;
        
        %% Spatial grids
        ny              = nx;
        Ly              = Lx;
        xvec            = linspace(-Lx/2, Lx/2, nx);
        yvec            = linspace(-Ly/2, Ly/2, ny);
        
        [ymat,xmat]     = meshgrid(yvec, xvec);
        dx              = Lx / (nx-1);
        dy              = Ly / (ny-1);
        %% Irradiance profiles
        
        I_red1_xy       = exp(-log(2) * ((2*xmat/diameters(1,1)).^2 + (2*ymat/diameters(1,2)).^2).^sg_coeffs(1));
        I_red2_xy       = exp(-log(2) * ((2*xmat/diameters(2,1)).^2 + (2*ymat/diameters(2,2)).^2).^sg_coeffs(2));
        I_blue_xy       = exp(-log(2) * ((2*(xmat-blue_offsets(1))/diameters(3,1)).^2 + (2*(ymat-blue_offsets(2))/diameters(3,2)).^2).^sg_coeffs(3));
        
        % Make irradiances have 1 watt power
        I_red1_xy_norm  = I_red1_xy ./ (dx*dy*trapz(trapz(I_red1_xy)));
        I_red2_xy_norm  = I_red2_xy ./ (dx*dy*trapz(trapz(I_red2_xy)));
        I_blue_xy_norm  = I_blue_xy ./ (dx*dy*trapz(trapz(I_blue_xy)));

        I_red1_xyt_M1   = repmat(permute(P.P_red1(:,1),[2,3,1]),[nx,ny,1]) .* repmat(I_red1_xy_norm,[1,1,nt]);
        I_red1_xyt_M2   = repmat(permute(P.P_red1(:,2),[2,3,1]),[nx,ny,1]) .* repmat(I_red1_xy_norm,[1,1,nt]);
        I_red2_xyt_M1   = repmat(permute(P.P_red2(:,1),[2,3,1]),[nx,ny,1]) .* repmat(I_red2_xy_norm,[1,1,nt]);
        I_red2_xyt_M2   = repmat(permute(P.P_red2(:,2),[2,3,1]),[nx,ny,1]) .* repmat(I_red2_xy_norm,[1,1,nt]);
        I_blue_xyt_M1   = repmat(permute(P.P_blue(:,1),[2,3,1]),[nx,ny,1]) .* repmat(I_blue_xy_norm,[1,1,nt]);
        % keyboard;
        O.I_red1_xy_norm = I_red1_xy_norm;
        O.I_red2_xy_norm = I_red2_xy_norm;
        O.I_blue_xy_norm = I_blue_xy_norm;

        O.I_red1_xyt_M1 = I_red1_xyt_M1;
        O.I_red1_xyt_M2 = I_red1_xyt_M2;
        O.I_red2_xyt_M1 = I_red2_xyt_M1;
        O.I_red2_xyt_M2 = I_red2_xyt_M2;
        O.I_blue_xyt_M1 = I_blue_xyt_M1;
        O.xmat          = xmat;
        O.ymat          = ymat;
        O.tvec_M1       = tvec_M1;
        O.tvec_M2       = tvec_M2;
        I_maxs = [max(O.I_red1_xyt_M1,[],'all'), max(O.I_red1_xyt_M2,[],'all');
            max(O.I_red2_xyt_M1,[],'all'), max(O.I_red2_xyt_M2,[],'all');
            max(O.I_blue_xyt_M1,[],'all'), 0];
        O.I_maxs        = I_maxs;
        O.pulses        = P;
        % keyboard;
    end
    
    % function [I_maxs, M_squareds, wxs] = superg_nums(obj)
    %     orders      = 1:10;
    %     wx_on_Xs    = [0.8493, 0.5821, 0.5344, 0.5176, 0.5099, 0.5058, 0.5034, 0.5019, 0.5009, 0.5003];
    %     M2_os       = [1, 1.128, 1.303, 1.469, 1.622, 1.764, 1.897, 2.022, 2.140, 2.252];
    %     S_M_Xsq_on_Pop = [0.8822, 1.196, 1.262, 1.282, 1.289, 1.291, 1.292, 1.291, 1.291, 1.290];
    %     I_maxs = zeros(3,2);
    %     if (obj.ristra_12_beam_diameters(1,1) ~= obj.ristra_12_beam_diameters(1,2))
    %         keyboard;
    %     end
    %     if (obj.ristra_12_beam_diameters(2,1) ~= obj.ristra_12_beam_diameters(2,2))
    %         keyboard;
    %     end
    %     if (obj.ristra_12_beam_diameters(3,1) ~= obj.ristra_12_beam_diameters(3,2))
        % keyboard;
    %     end
    %     Xs  = obj.ristra_12_beam_diameters(:,1);
    %     Xs  = Xs*1e-3; % mm --> m
    %     P   = obj.make_pulses;
    %     % keyboard;
    %     spatial_sg  = obj.ristra_12_spatial_supergaussian;
    %     I_maxs      = zeros(3,2);
    %     M_squareds  = zeros(3,1);
    %     wx_on_X     = spline(orders, wx_on_Xs, spatial_sg);
    %     wxs         = wx_on_X .* Xs;
    %     M_squareds  = spline(orders, M2_os, spatial_sg);
    %     I_maxs      = zeros(3,2);
    %     Pmax = max(P.P_red1,[],1);
    %     I_maxs(1,1) = spline(orders, S_M_Xsq_on_Pop, spatial_sg(1)) / Xs(1)^2 * Pmax(1);
    %     I_maxs(1,2) = spline(orders, S_M_Xsq_on_Pop, spatial_sg(1)) / Xs(1)^2 * Pmax(2);
    %     Pmax = max(P.P_red2,[],1);
    %     I_maxs(2,1) = spline(orders, S_M_Xsq_on_Pop, spatial_sg(2)) / Xs(2)^2 * Pmax(1);
    %     I_maxs(2,2) = spline(orders, S_M_Xsq_on_Pop, spatial_sg(2)) / Xs(2)^2 * Pmax(2);
    %     Pmax = max(P.P_blue,[],1);
    %     I_maxs(3,1) = spline(orders, S_M_Xsq_on_Pop, spatial_sg(3)) / Xs(3)^2 * Pmax(1);
    % end

    function show(obj)
        % Function to display the values of properties of  this object in
        % command window; if custom input fields are provided to ristra12 model
        % many of these values will not be used - ristra_12_pulse_duration,
        % ristra_12_temporal_sg, ristra_12_pulse_delay,
        % ristra_12_beam_diameters, ristra_12_spatial_supergaussian, ristra_12_blue_beam_offset_xy,
        % ristra_12_beam_roc, ristra_12_nxy, and ristra_12_lxy.
        fprintf(1, 'ristra_12_wavelengths (nm):         %g, %g, %g\n', obj.ristra_12_wavelengths);
        fprintf(1, 'ristra_12_ref_inds:                 %g, %g, %g\n', obj.ristra_12_ref_inds);
        fprintf(1, 'ristra_12_left_refl_c:              %g, %g, %g\n', obj.ristra_12_left_refl_c);
        fprintf(1, 'ristra_12_right_refl_c:             %g, %g, %g\n', obj.ristra_12_right_refl_c);
        fprintf(1, 'ristra_12_crystal_losss (per mm):   %g, %g, %g\n', obj.ristra_12_crystal_loss);
        fprintf(1, 'ristra_12_left_energy_pwr (J or W): %g, %g, %g\n', obj.ristra_12_left_energy_pwr);
        fprintf(1, 'ristra_12_right_energy_pwr (J or W):%g, %g, %g\n', obj.ristra_12_right_energy_pwr);
        fprintf(1, 'ristra_12_pulse_duration (ns):      %g, %g, %g\n', obj.ristra_12_pulse_duration);
        fprintf(1, 'ristra_12_temporal_sg:              %g, %g, %g\n', obj.ristra_12_temporal_sg);
        fprintf(1, 'ristra_12_pulse_delay (ns):         %g, %g, %g\n', obj.ristra_12_pulse_delay(1), obj.ristra_12_pulse_delay(2), 0);
        fprintf(1, 'ristra_12_beam_diameters (x) (mm):  %g, %g, %g\n', obj.ristra_12_beam_diameters(:,1));
        fprintf(1, 'ristra_12_beam_diameters (y) (mm):  %g, %g, %g\n', obj.ristra_12_beam_diameters(:,2));
        fprintf(1, 'ristra_12_spatial_supergaussian:    %g, %g, %g\n', obj.ristra_12_spatial_supergaussian);
        fprintf(1, 'ristra_12_blue_beam_offset_xy (mm): %g, %g\n', obj.ristra_12_blue_beam_offset_xy);
        fprintf(1, 'ristra_12_walkoff_angles (mrad):    %g, %g, %g\n', obj.ristra_12_walkoff_angles);
        fprintf(1, 'ristra_12_beam_roc (x) (mm):        %g, %g, %g\n', obj.ristra_12_beam_roc(:,1));
        fprintf(1, 'ristra_12_beam_roc (y) (mm):        %g, %g, %g\n', obj.ristra_12_beam_roc(:,2));
        fprintf(1, 'ristra_12_left_refl_m:              %g, %g, %g\n', obj.ristra_12_left_refl_m);
        fprintf(1, 'ristra_12_right_refl_m:             %g, %g, %g\n', obj.ristra_12_right_refl_m);
        fprintf(1, 'ristra_12_phase_1 (rad):            %g, %g, %g\n', obj.ristra_12_phase_1);
        fprintf(1, 'ristra_12_phase_2 (rad):            %g, %g, %g\n', obj.ristra_12_phase_2);
        fprintf(1, 'ristra_12_nz:                       %i\n', obj.ristra_12_nz);
        fprintf(1, 'ristra_12_crystal_lengths (mm):     %g, %g\n', obj.ristra_12_crystal_lengths);
        fprintf(1, 'ristra_12_cavity_length (mm):       %g\n', obj.ristra_12_cavity_length);
        fprintf(1, 'ristra_12_nxy:                      %i\n', obj.ristra_12_nxy);
        fprintf(1, 'ristra_12_lxy (mm):                 %g\n', obj.ristra_12_lxy);
        fprintf(1, 'ristra_12_deff (pm/V):              %g\n', obj.ristra_12_deff);
        fprintf(1, 'ristra_12_deltak (per mm):          %g\n', obj.ristra_12_deltak);
        fprintf(1, 'ristra_12_nstarts:                  %i\n', obj.ristra_12_nstarts);
        fprintf(1, 'ristra_12_add_quantum_noise:        %i\n', obj.ristra_12_add_quantum_noise);
        fprintf(1, 'ristra_12_disable_image_rotation:   %i\n', obj.ristra_12_disable_image_rotation);
    end
end
end


% The model starts with the first point in time at the left mirror (also
% called mirror 1, or input mirror), enters the cavity by transmitting
% through mirror 1, propagates to the input face of crystal 1, enters
% crystal 1, mixes and propagates through crystal 1, exits crystal,
% propagates to mirror 2, reflect & transmit through mirror 2, through the
% first half of leg 2 to crystal 2, enters crystal 2, mixes and propagates
% through crystal 2, exits crystal 2, propagates last half of leg 2 to the
% input mirror, and reflects & transmits through mirror 1; now step time &
% move to the next cavity roundtrip.

% At both the input and output mirror, electric fields for each wave enter &
% exit cavity, with the exception of having no input field for the blue wave
% at the output mirror.
%
% For long cavities or short pulses, the time step (set by the cavity
% roundtrip time) may be too large, in which case you can use multiple
% 'starts.'